#define T_EOL '\n'
#define T_BREAK 256
#define T_TEMP 257
#define T_BT 258
#define T_CLEAR 259
#define T_CB 260
#define T_LIST 261
#define T_LB 262
#define T_NEXT 263
#define T_STEP 264
#define T_ASGN 265
#define T_GO 266
#define T_DUMP 267
#define T_REGISTERS 268
#define T_DR 269
#define T_MEMORY 270
#define T_DM 270
#define T_IO 271
#define T_LI 272
#define T_QUIT 273
#define T_CONT 274
#define T_UNASM 275
#define T_NUMBER 276
#define T_STRING 277 \

#define MAX_TOKENSIZE 256 \

/*3:*/
#line 74 "./mixterm.w"

#ifdef HAVE_CONFIG_H
# include <config.h> 
#endif
#include <stdlib.h> 
#include <stdio.h> 
#include <ctype.h> 
#include <string.h> 
#include <errno.h> 
#include <mix.h> 
#include <mixsim.h> 
#ifdef WITH_READLINE
# include <readline/readline.h> 
#endif

/*:3*//*4:*/
#line 90 "./mixterm.w"

struct command_word{
char*longname;
char*shortname;
int tok;
}cword[]= {
{"BREAK","B",T_BREAK},
{"TEMP",NULL,T_TEMP},
{"BT",NULL,T_BT},
{"CLEAR",NULL,T_CLEAR},
{"CB",NULL,T_CB},
{"LIST",NULL,T_LIST},
{"LB",NULL,T_LB},
{"NEXT","N",T_NEXT},
{"STEP","S",T_STEP},
{"ASGN","A",T_ASGN},
{"GO","G",T_GO},
{"DUMP","DU",T_DUMP},
{"REGISTERS","REG",T_REGISTERS},
{"DR",NULL,T_DR},
{"MEMORY","MEM",T_MEMORY},
{"DM",NULL,T_DM},
{"IO",NULL,T_IO},
{"LI",NULL,T_LI},
{"CONT","CO",T_CONT},
{"UNASM","U",T_UNASM},
{"QUIT","Q",T_QUIT},
{0}
};

/*:4*//*5:*/
#line 121 "./mixterm.w"

struct{
int type;
int number;
char string[MAX_TOKENSIZE];
}token;
char*buffer;
char*curp;
char prompt[]= "MIX> ";
int interactive;


/*:5*//*6:*/
#line 140 "./mixterm.w"

#ifdef WITH_READLINE

/*:6*//*7:*/
#line 146 "./mixterm.w"

char*
mixterm_command_generator(const char*text,int state)
{
static int i,len;
const char*name;

if(!state){
i= 0;
len= strlen(text);
}

while((name= cword[i].longname)){
if(cword[i].shortname&&strlen(cword[i].shortname)> strlen(name))
name= cword[i].shortname;
i++;
if(strncasecmp(name,text,len)==0)
return strdup(name);
}

return NULL;
}

/*:7*//*8:*/
#line 173 "./mixterm.w"

char**
mixterm_command_completion(char*cmd,int start,int end)
{
if(start==0)
return rl_completion_matches(cmd,mixterm_command_generator);
return NULL;
}

/*:8*//*9:*/
#line 183 "./mixterm.w"

static int
mixterm_getc(FILE*stream)
{
unsigned char c;

while(1){
if(read(0,&c,1)==1)
return c;
if(errno==EINTR)
break;
}
return EOF;
}

/*:9*//*10:*/
#line 199 "./mixterm.w"

void
mixterm_readline_init()
{
if(!interactive)
return;

rl_readline_name= "mixsim";
rl_attempted_completion_function= (CPPFunction*)mixterm_command_completion;
rl_getc_function= mixterm_getc;
}

/*:10*//*11:*/
#line 212 "./mixterm.w"

#endif

/*:11*//*12:*/
#line 217 "./mixterm.w"

char*
mixterm_readline_internal()
{
char*line;
char*p;
size_t alloclen,linelen;

p= line= calloc(1,255);
if(!p)
abort();
alloclen= 255;
linelen= 0;
for(;;){
size_t n;

p= fgets(p,alloclen-linelen,stdin);

if(p)
n= strlen(p);
else{
free(line);
return NULL;
}

linelen+= n;


if(linelen==0){
free(line);
return NULL;
}


if(line[linelen-1]=='\n'){
line[linelen-1]= '\0';
return line;
}else{
char*tmp;
alloclen*= 2;
tmp= realloc(line,alloclen);
if(tmp==NULL)
abort();
line= tmp;
p= line+linelen;
}
}
}

/*:12*//*13:*/
#line 267 "./mixterm.w"

char*
mixterm_readline(char*prompt)
{
if(interactive)
return readline(prompt);
return mixterm_readline_internal();
}

/*:13*//*14:*/
#line 277 "./mixterm.w"

#ifndef WITH_READLINE
char*
readline(char*prompt)
{
if(prompt){
printf("%s",prompt);
fflush(stdout);
}

return mixterm_readline_internal();
}

#define mixterm_readline_init()
#endif


/*:14*//*16:*/
#line 297 "./mixterm.w"

void
copy_token()
{
char*p;

p= token.string;
while(*curp&&!isspace(*curp))
*p++= *curp++;
*p= 0;
}

/*:16*//*17:*/
#line 312 "./mixterm.w"

int
nextkn()
{
if(!curp){
char*p;

p= curp= readline(prompt);
if(curp){
while(*curp&&(*curp==' '||*curp=='\t'))
curp++;
if(*curp==0)

curp= buffer;
else{
free(buffer);
buffer= p;
#ifdef WITH_READLINE
add_history(curp);
#endif
}
}
}

if(!curp)
return 0;

while(*curp&&(*curp==' '||*curp=='\t'))
curp++;

if(!*curp){
curp= NULL;
return token.type= T_EOL;
}

if(isalpha(*curp)||*curp=='/'){
struct command_word*cwp;

copy_token();
for(cwp= cword;cwp->longname;cwp++){
if(strcasecmp(token.string,cwp->longname)==0
||(cwp->shortname
&&strcasecmp(token.string,cwp->shortname)==0))
return token.type= cwp->tok;

}
return token.type= T_STRING;
}
if(isdigit(*curp)){
copy_token();
token.number= strtol(token.string,NULL,0);
return token.type= T_NUMBER;
}
return token.type= *curp++;
}

/*:17*//*18:*/
#line 370 "./mixterm.w"

int
getaddr()
{
int addr= token.number;

if(addr<0||addr>=MEMSIZE){
printf("LOCATION OUT OF ADDRESS SPACE\n");
return-1;
}
return addr;
}

/*:18*//*19:*/
#line 406 "./mixterm.w"

void
m_break()
{
int addr;
int count= -1;

if(nextkn()==T_TEMP){
count= 1;
nextkn();
}
if(token.type!=T_NUMBER){
printf("USAGE: BREAK [TEMP] <ADDR> [<COUNT>]\n");
return;
}
if((addr= getaddr())==-1)
return;
nextkn();
if(count==-1&&token.type==T_NUMBER){
count= token.number;
nextkn();
}
if(token.type!=T_EOL){
printf("USAGE: BREAK [TEMP] <ADDR> [<COUNT>]\n");
return;
}
set_bp(addr,count);
}

/*:19*//*20:*/
#line 436 "./mixterm.w"

void
m_bt()
{
int addr;

if(nextkn()!=T_NUMBER){
printf("USAGE: BT <ADDR>\n");
return;
}
if((addr= getaddr())==-1)
return;
set_bp(addr,1);
}

/*:20*//*21:*/
#line 460 "./mixterm.w"

void
m_clear()
{
int addr;

if(nextkn()!=T_BREAK){
printf("EXPECTED BREAK\n");
return;
}
if(nextkn()!=T_NUMBER
||(addr= getaddr())==-1
||nextkn()!=T_EOL){
printf("USAGE: CLEAR BREAK <ADDR>\n");
return;
}
clear_bp(addr);
}

/*:21*//*22:*/
#line 482 "./mixterm.w"

void
m_cb()
{
int addr;

if(nextkn()!=T_NUMBER
||(addr= getaddr())==-1
||nextkn()!=T_EOL){
printf("USAGE: CB <ADDR>\n");
return;
}
clear_bp(addr);
}

/*:22*//*23:*/
#line 498 "./mixterm.w"

void
m_lb()
{
if(nextkn()!=T_EOL){
printf("EOL MISSING\n");
return;
}
list_bp();
}

/*:23*//*24:*/
#line 510 "./mixterm.w"

void
m_lio()
{
int dev= -1;

if(nextkn()==T_NUMBER){
dev= token.number;
nextkn();
}
if(token.type!=T_EOL){
printf("EOL MISSING\n");
return;
}
list_io(dev);
}

/*:24*//*25:*/
#line 536 "./mixterm.w"

void
m_list()
{
switch(nextkn()){
case T_IO:
m_lio();
break;
case T_BREAK:
m_lb();
break;
default:
printf("NOT RECOGNIZED\n");
}
}

/*:25*//*26:*/
#line 553 "./mixterm.w"

void
m_next()
{
nextkn();
if(token.type!=T_EOL){
printf("EOL MISSING\n");
return;
}
set_next(F);
run();
}

/*:26*//*27:*/
#line 567 "./mixterm.w"

void
m_step()
{
if(nextkn()!=T_EOL){
printf("EOL MISSING\n");
return;
}
set_next(T);
run();
}

/*:27*//*28:*/
#line 580 "./mixterm.w"

void
m_cont()
{
if(nextkn()!=T_EOL){
printf("EOL MISSING\n");
return;
}
run();
}

/*:28*//*29:*/
#line 592 "./mixterm.w"

void
m_asgn()
{
int devno;

if(nextkn()!=T_NUMBER){
printf("ASGN <DEV> <UNIX-PATH>\n");
return;
}
devno= token.number;
if(nextkn()==T_EOL){
printf("ASGN <DEV> <UNIX-PATH>\n");
return;
}
asgn_io(devno,token.string);
}

/*:29*//*30:*/
#line 611 "./mixterm.w"

void
m_go()
{
if(nextkn()!=T_EOL){
printf("EOL MISSING\n");
return;
}
go();
}

/*:30*//*31:*/
#line 623 "./mixterm.w"

void
m_dr()
{
if(nextkn()!=T_EOL){
printf("EOL MISSING\n");
return;
}
dump_status();
}

/*:31*//*32:*/
#line 642 "./mixterm.w"

void
m_dm()
{
int start_loc= 0,end_loc= MEMSIZE;

if(nextkn()==T_NUMBER){
if((start_loc= getaddr())==-1){
return;
}
if(nextkn()==T_NUMBER){
if((end_loc= getaddr())==-1)
return;
nextkn();
}
}
if(end_loc<start_loc){
printf("BAD RANGE\n");
return;
}
if(token.type!=T_EOL){
printf("EOL MISSING\n");
return;
}
dump_memory(start_loc,end_loc);
}

/*:32*//*33:*/
#line 670 "./mixterm.w"

void
m_dump()
{
switch(nextkn()){
case T_MEMORY:
m_dm();
break;
case T_REGISTERS:
m_dr();
break;
default:
dump();
break;
}
}

/*:33*//*34:*/
#line 692 "./mixterm.w"

void
m_unasm()
{
extern int loc;
int start_loc= loc,end_loc= loc+5;

if(nextkn()==T_NUMBER){
if((start_loc= getaddr())==-1)
return;
if(nextkn()==T_NUMBER){
if((end_loc= getaddr())==-1)
return;
nextkn();
}else
end_loc= start_loc+5;
}
if(end_loc<start_loc){
printf("BAD RANGE\n");
return;
}
if(token.type!=T_EOL){
printf("EOL MISSING\n");
return;
}
disas(stdout,start_loc,end_loc);
}

/*:34*//*35:*/
#line 721 "./mixterm.w"

void
m_quit()
{
exit(0);
}

/*:35*//*36:*/
#line 731 "./mixterm.w"

void
m_eol()
{
}

/*:36*//*37:*/
#line 738 "./mixterm.w"

void
m_err()
{
printf("NOT RECOGNIZED\n");
curp= NULL;
}

/*:37*//*38:*/
#line 747 "./mixterm.w"

struct disp{
int tok;
void(*fun)(void);
}main_disp[]= {
{T_BREAK,m_break},
{T_BT,m_bt},
{T_CLEAR,m_clear},
{T_CB,m_cb},
{T_LIST,m_list},
{T_LB,m_lb},
{T_NEXT,m_next},
{T_STEP,m_step},
{T_ASGN,m_asgn},
{T_GO,m_go},
{T_DUMP,m_dump},
{T_DR,m_dr},
{T_DM,m_dm},
{T_LIST,m_list},
{T_LI,m_lio},
{T_QUIT,m_quit},
{T_EOL,m_eol},
{T_CONT,m_cont},
{T_UNASM,m_unasm},
{0}
};

/*:38*//*39:*/
#line 775 "./mixterm.w"

char init_prompt[]= "\nMIX TERMINAL STATION READY FOR COMMUNICATION\n";

/*:39*//*40:*/
#line 779 "./mixterm.w"

void
mixterm()
{
struct disp*dsp;
void(*fun)();

interactive= isatty(0);
mixterm_readline_init();
printf("%s\n",init_prompt);
while(nextkn()){
fun= m_err;
for(dsp= main_disp;dsp->tok;dsp++){
if(token.type==dsp->tok){
fun= dsp->fun;
break;
}
}
(*fun)();
}
}

/*:40*/
